#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Hamilton Kibbe <ham@hamiltonkib.be>

from dataclasses import dataclass
from typing import NamedTuple

from tutti.base import LockConfig, SemaphoreConfig


class RedisSemaphoreHandle(NamedTuple):
    name: str
    identifier: str


@dataclass
class RedisLockConfig(LockConfig):
    """Lock congfiguration Base class.

    Parameters:
        name (str):
            The name of the Redis Lock.
        timeout (float):
            The timeout for acquiring the lock, in seconds
        blocking (bool):
            Whether to block while waiting for the lock to be released.
        connection_url (str):
            The Redis connection URL.
        blocking (bool):
            Whether to block while waiting for the lock to be released. Defaults to `True`.
    """
    connection_url: str


@dataclass
class RedisSemaphoreConfig(SemaphoreConfig):
    """Configuration for Redis semaphores.

    Parameters:
        name (str):
            The name of the Semaphore.
        max_concurrency (int):
            The maximum number of concurrent accesses to the resource.
        connection_url (str):
            The Redis connection URL.
        lock_timeout (float):
            The timeout for acquiring the lock, in seconds
        lock_blocking (bool):
            Whether to block while waiting for the lock to be released. Defaults to `True`.
    """
    connection_url: str
    lock_timeout: float



from unittest.mock import Mock

from redis import Redis
from redis.lock import Lock as RedisLock
import pytest

from tutti.backends.redis.sync import Lock, Semaphore, RedisWrapper
from tutti.backends.redis.types import RedisSemaphoreHandle


# region mocks

class MockRedisLock(RedisLock):
    def __init__(self, *args, **kwargs):
        pass


class MockRedisConnection(Redis):
    def __init__(self, *args, **kwargs):
        pass


# endregion

def test_lock_success():
    """
    Given:
        - A RedisWraper mock that executes correctly
        - A Tutti redis Lock

    When:
        - The lock is acquired
        - The lock is released

    Then:
        - The lock is acquired and released successfully
    """
    # Given
    mock_redis_lock = MockRedisLock()
    mock_redis_wrapper = Mock(spec=RedisWrapper)
    mock_redis_wrapper.acquire_lock.return_value = (mock_redis_lock, True)
    mock_redis_wrapper.release_lock.return_value = True

    # When
    with Lock(
        connection_url="shouldn't_matter",
        lock_name="test_lock",
        timeout=10,
        conn=MockRedisConnection(),
        redis_wrapper=mock_redis_wrapper  # type: ignore
    ) as lock:
        # Then
        assert lock._handle == mock_redis_lock

    mock_redis_wrapper.acquire_lock.assert_called_once()
    mock_redis_wrapper.release_lock.assert_called_once()


def test_lock_failure_to_acquire():
    """
    Given:
        - A RedisWraper mock that does not azquire the lock
        - A Tutti redis Lock

    When:
        - The lock is acquired
        - The lock is released

    Then:
        - A RuntimeError error is raised
        - acquire_lock is called
        - release_lock is not called
    """
    # Given
    mock_redis_lock = MockRedisLock()
    mock_redis_wrapper = Mock(spec=RedisWrapper)
    mock_redis_wrapper.acquire_lock.return_value = (mock_redis_lock, False)

    # When
    with pytest.raises(RuntimeError):
        with Lock(
            connection_url="shouldn't_matter",
            lock_name="test_lock",
            timeout=10,
            conn=MockRedisConnection(),
            redis_wrapper=mock_redis_wrapper  # type: ignore
        ) as lock:
            # Then
            assert lock._handle is None

    mock_redis_wrapper.acquire_lock.assert_called_once()
    mock_redis_wrapper.release_lock.assert_not_called()


def test_lock_raises():
    """
    Given:
        - A RedisWraper mock that raises when acquiring the lock
        - A Tutti redis Lock
    
    When:
        - The lock is acquired
        - The lock is released
    
    Then:
        - A RuntimeError error is raised
        - acquire_lock is called
        - release_lock is not called
    """
    # Given
    mock_redis_wrapper = Mock(spec=RedisWrapper)
    mock_redis_wrapper.acquire_lock.side_effect = ValueError("something happened")

    # When
    with pytest.raises(RuntimeError):
        with Lock(
            connection_url="shouldn't_matter",
            lock_name="test_lock",
            timeout=10,
            conn=MockRedisConnection(),
            redis_wrapper=mock_redis_wrapper  # type: ignore
        ) as lock:
            # Then
            assert lock._handle is None

    mock_redis_wrapper.acquire_lock.assert_called_once()
    mock_redis_wrapper.release_lock.assert_not_called()


def test_semaphore_success():
    """
    Given:
        - A RedisWraper mock that executes correctly
        - A Tutti redis Semaphore

    When:
        - The semaphore is acquired
        - The semaphore is released

    Then:
        - The semaphore is acquired and released successfully
    """
    # Given
    semaphore_handle = RedisSemaphoreHandle("test_lock", "whatever")
    mock_redis_lock = MockRedisLock()
    mock_redis_wrapper = Mock(spec=RedisWrapper)
    mock_redis_wrapper.acquire_lock.return_value = (mock_redis_lock, True)
    mock_redis_wrapper.release_lock.return_value = True
    mock_redis_wrapper.acquire_semaphore.return_value = semaphore_handle
    mock_redis_wrapper.release_semaphore.return_value = True

    # When
    with Semaphore(
        connection_url="shouldn't_matter",
        lock_name="test_lock",
        value=1,
        timeout=10,
        conn=MockRedisConnection(),
        redis_wrapper=mock_redis_wrapper  # type: ignore
    ) as semaphore:
        # Then
        assert semaphore._handle == semaphore_handle

    mock_redis_wrapper.acquire_semaphore.assert_called_once()
    mock_redis_wrapper.release_semaphore.assert_called_once()


def test_semaphore_fail_to_acquire():
    """
    Given:
        - A RedisWraper mock that fails to acquire the semaphore
        - A Tutti redis Semaphore

    When:
        - The semaphore is acquired
        - The semaphore is released

    Then:
        - A RuntimeError error is raised
        - acquire_semaphore is called
        - release_semaphore is not called
    """
    # Given
    mock_redis_lock = MockRedisLock()
    mock_redis_wrapper = Mock(spec=RedisWrapper)
    mock_redis_wrapper.acquire_lock.return_value = (mock_redis_lock, True)
    mock_redis_wrapper.release_lock.return_value = True
    mock_redis_wrapper.acquire_semaphore.return_value = None

    # When
    with pytest.raises(RuntimeError):
        with Semaphore(
            connection_url="shouldn't_matter",
            lock_name="test_lock",
            value=1,
            timeout=10,
            conn=MockRedisConnection(),
            redis_wrapper=mock_redis_wrapper  # type: ignore
        ) as semaphore:
            # Then
            assert semaphore._handle is None

    mock_redis_wrapper.acquire_semaphore.assert_called_once()
    mock_redis_wrapper.release_semaphore.not_called()


def test_semaphore_fail_to_acquire_lock():
    """
    Given:
        - A RedisWraper mock that raises when acquiring the lock
        - A Tutti redis Semaphore

    When:
        - The semaphore is acquired
        - The semaphore is released

    Then:
        - A RuntimeError error is raised
        - acquire_semaphore is not called
        - release_semaphore is not called
    """
    # Given
    mock_redis_wrapper = Mock(spec=RedisWrapper)
    mock_redis_wrapper.acquire_lock.side_effect = ValueError("something happened")

    # When
    with pytest.raises(RuntimeError):
        with Semaphore(
            connection_url="shouldn't_matter",
            lock_name="test_lock",
            value=1,
            timeout=10,
            conn=MockRedisConnection(),
            redis_wrapper=mock_redis_wrapper  # type: ignore
        ) as semaphore:
            # Then
            assert semaphore._handle is None

    mock_redis_wrapper.acquire_semaphore.not_called()
    mock_redis_wrapper.release_semaphore.not_called()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Hamilton Kibbe <ham@hamiltonkib.be>
#
import asyncio

from colorama import Fore, Style

from tutti.configuration import RedisLockConfig, RedisSemaphoreConfig
from tutti.asyncio import Lock, Semaphore


CONNNECTION_URL = "redis://localhost:6379/0"

REDIS_LOCK_CONFIG = RedisLockConfig(
    connection_url=CONNNECTION_URL,
    name="demo-lock",
    timeout=5,
)

REDIS_SEMAPHORE_CONFIG = RedisSemaphoreConfig(
    connection_url=CONNNECTION_URL,
    max_concurrency=2,
    name="demo-semaphore",
    lock_timeout=5,
)


def pprint(data, task_number):
    color = [Fore.CYAN, Fore.MAGENTA, Fore.YELLOW, Fore.GREEN, Fore.BLUE, Fore.RED][task_number]
    print(f"{color}    {data}{Style.RESET_ALL}")


async def use_protected_resource():
    # Just pretend...
    await asyncio.sleep(0.3)


async def access_exclusive_resource(task_number: int) -> None:
    """Use a distributed lock to limit access to a critical resource"""

    async with Lock(REDIS_LOCK_CONFIG):
        pprint(f"Task {task_number} Entering critical section", task_number)
        await use_protected_resource()
        pprint(f"Task {task_number} Leaving critical section", task_number)


async def access_limited_resource(task_number: int) -> None:
    """Use a distributed semaphore to limit access to a critical resource"""

    async with Semaphore(REDIS_SEMAPHORE_CONFIG):
        pprint(f"Task {task_number} Entering critical section", task_number)
        await use_protected_resource()
        pprint(f"Task {task_number} Leaving critical section", task_number)


async def main():
    print("\n============================================================================")
    print("Lock demo: Only one task may be in the critical section at a time")
    print("============================================================================")
    lock_tasks = [
        asyncio.create_task(access_exclusive_resource(i)) for i in range(4)
    ]
    await asyncio.gather(*lock_tasks)

    print("\n============================================================================")
    print("Semaphore demo: Up to two tasks may be in the critical section at a time")
    print("============================================================================")
    semaphore_tasks = [
        asyncio.create_task(access_limited_resource(i)) for i in range(4)
    ]
    await asyncio.gather(*semaphore_tasks)


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Hamilton Kibbe <ham@hamiltonkib.be>

import asyncio
import time
import uuid

from redis import Redis
from redis.lock import Lock as RedisLock
from redis.asyncio import Redis as AsyncRedis
from redis.asyncio.lock import Lock as AsyncRedisLock

from .types import RedisSemaphoreHandle


def acquire_lock(
    conn: Redis,
    lock_name: str,
    blocking: bool = True,
    timeout: float | None = None
) -> tuple[RedisLock, bool]:
    lock = conn.lock(
        lock_name,
        timeout=timeout,
        blocking_timeout=None if not blocking else timeout
    )
    result = lock.acquire()
    return lock, result


async def aacquire_lock(
    conn: AsyncRedis,
    lock_name: str,
    blocking: bool = True,
    timeout: float | None = None
) -> tuple[AsyncRedisLock, bool]:
    lock = conn.lock(
        lock_name,
        timeout=timeout,
        blocking_timeout=None if not blocking else timeout
    )
    result = await lock.acquire()
    return lock, result


def release_lock(conn: Redis, lock: RedisLock) -> bool:
    lock.release()
    return True


async def arelease_lock(conn: AsyncRedis, lock: AsyncRedisLock) -> bool:
    await lock.release()
    return True


def acquire_semaphore(
    conn: Redis,
    lock_name: str,
    value: int = 1,
    blocking: bool = True,
    timeout: float = -1
) -> RedisSemaphoreHandle | None:

    identifier = str(uuid.uuid4())
    czset = f"{lock_name}-owner"
    ctr = f"{lock_name}-counter"

    now = time.time()

    while True:
        pipeline = conn.pipeline(transaction=True)

        if timeout >= 0:
            pipeline.zremrangebyscore(lock_name, "-inf", now - timeout)

        pipeline.zinterstore(czset, {czset: 1, lock_name: 0})
        pipeline.incr(ctr)
        counter = pipeline.execute()[-1]
        pipeline.zadd(lock_name, {identifier: now})
        pipeline.zadd(czset, {identifier: counter})
        pipeline.zrank(czset, identifier)
        result = pipeline.execute()
        if result[-1] < value:
            return RedisSemaphoreHandle(lock_name, identifier)
        pipeline.zrem(lock_name, identifier)
        pipeline.zrem(czset, identifier)
        pipeline.execute()

        if (not blocking) or (timeout >= 0 and time.time() > (now + timeout)):
            return None
        time.sleep(0.001)


async def aacquire_semaphore(
    conn: AsyncRedis,
    lock_name: str,
    value: int = 1,
    blocking: bool = True,
    timeout: float = -1
) -> RedisSemaphoreHandle | None:

    identifier = str(uuid.uuid4())
    czset = f"{lock_name}-owner"
    ctr = f"{lock_name}-counter"

    now = time.time()

    while True:
        pipeline = conn.pipeline(transaction=True)

        if timeout >= 0:
            pipeline.zremrangebyscore(lock_name, "-inf", now - timeout)

        pipeline.zinterstore(czset, {czset: 1, lock_name: 0})
        pipeline.incr(ctr)
        counter_coroutine = await pipeline.execute()
        counter = counter_coroutine[-1]
        pipeline.zadd(lock_name, {identifier: now})
        pipeline.zadd(czset, {identifier: counter})
        pipeline.zrank(czset, identifier)
        result = await pipeline.execute()
        if result[-1] < value:
            return RedisSemaphoreHandle(lock_name, identifier)
        pipeline.zrem(lock_name, identifier)
        pipeline.zrem(czset, identifier)
        await pipeline.execute()

        if (not blocking) or (timeout >= 0 and time.time() > (now + timeout)):
            return None
        await asyncio.sleep(0.001)


def release_semaphore(conn: Redis, lock: RedisSemaphoreHandle) -> bool:
    pipeline = conn.pipeline(transaction=True)
    pipeline.zrem(lock.name, lock.identifier)
    pipeline.zrem(f"{lock.name}-owner", lock.identifier)
    result = pipeline.execute()
    return bool(result[0])


async def arelease_semaphore(conn: AsyncRedis, lock: RedisSemaphoreHandle) -> bool:
    pipeline = conn.pipeline(transaction=True)
    pipeline.zrem(lock.name, lock.identifier)
    pipeline.zrem(f"{lock.name}-owner", lock.identifier)
    result = await pipeline.execute()
    return bool(result[0])


def locked(conn: Redis, lock_name: str) -> bool:
    lock = conn.lock(lock_name)
    return lock.locked()


async def alocked(conn: AsyncRedis, lock_name: str) -> bool:
    lock = conn.lock(lock_name)
    return await lock.locked()



import pytest

from tutti import (
    Lock,
    Semaphore,
    BoundedSemaphore,
)
from tutti.configuration import RedisLockConfig, RedisSemaphoreConfig


from tutti.backends.redis import (
    Lock as RedisLock,
    Semaphore as RedisSemaphore,
    BoundedSemaphore as RedisBoundedSemaphore,
)

# region mocks

FAKE_CONNECTION_URL = "redis://notarealplace:6379/0"


# endregion


def test_redis_lock_factory():
    """
    Given:
        - A RedisLockConfig object
    When:
        - The lock factory is called with the config object
    Then:
        - The correct RedisLock instance is created
    """
    # Given
    config = RedisLockConfig(
        connection_url=FAKE_CONNECTION_URL,
        name="test_lock",
        timeout=10,
        blocking=True,
    )

    # When
    lock = Lock(config)

    # Then
    assert isinstance(lock, RedisLock)


def test_redis_semaphore_factory():
    """
    Given:
        - A RedisSemaphoreConfig object
    When:
        - The semaphore factory is called with the config object
    Then:
        - The correct RedisSemaphore instance is created
    """
    # Given
    config = RedisSemaphoreConfig(
        connection_url=FAKE_CONNECTION_URL,
        max_concurrency=5,
        name="test_semaphore",
        lock_timeout=10,
    )

    # When
    semaphore = Semaphore(config)

    # Then
    assert isinstance(semaphore, RedisSemaphore)


def test_redis_bounded_semaphore_factory():
    """
    Given:
        - A RedisSemaphoreConfig object
    When:
        - The bounded semaphore factory is called with the config object
    Then:
        - The correct RedisBoundedSemaphore instance is created
    """
    # Given
    config = RedisSemaphoreConfig(
        connection_url=FAKE_CONNECTION_URL,
        max_concurrency=5,
        name="test-semaphore",
        lock_timeout=10,
    )

    # When
    bounded_semaphore = BoundedSemaphore(config)

    # Then
    assert isinstance(bounded_semaphore, RedisBoundedSemaphore)


def test_redis_lock_factory_not_implemented():
    """
    Given:
        - A random class
        - An instance of the random class used as the config
    When:
        - The lock factory is called with the config object
    Then:
        - NotImplementedError is raised
    """
    # Given
    class RandomClass:
        pass
    config = RandomClass()

    # When/Then
    with pytest.raises(NotImplementedError):
        Lock(config)  # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Hamilton Kibbe <ham@hamiltonkib.be>
#

import multiprocessing
import time

from colorama import Fore, Style

from tutti import Lock, Semaphore
from tutti.configuration import RedisLockConfig, RedisSemaphoreConfig


CONNNECTION_URL = "redis://localhost:6379/0"

REDIS_LOCK_CONFIG = RedisLockConfig(
    connection_url=CONNNECTION_URL,
    name="demo-lock",
    timeout=5,
    blocking=True,
)

REDIS_SEMAPHORE_CONFIG = RedisSemaphoreConfig(
    connection_url=CONNNECTION_URL,
    max_concurrency=2,
    name="demo-semaphore",
    lock_timeout=5,
)


def pprint(data, process_id):
    color = [Fore.CYAN, Fore.MAGENTA, Fore.YELLOW, Fore.GREEN, Fore.BLUE, Fore.RED][process_id]
    print(f"{color}    {data}{Style.RESET_ALL}")


def use_protected_resource():
    # Just pretend...
    time.sleep(0.3)


def access_exclusive_resource(process_id: int) -> None:
    """Use a distributed lock to limit access to a critical resource"""

    with Lock(REDIS_LOCK_CONFIG):
        pprint(f"Process {process_id} Entering critical section", process_id)
        use_protected_resource()
        pprint(f"Process {process_id} Leaving critical section", process_id)


def access_limited_resource(process_id: int) -> None:
    """Use a distributed semaphore to limit access to a critical resource"""

    with Semaphore(REDIS_SEMAPHORE_CONFIG):
        pprint(f"Process {process_id} Entering critical section", process_id)
        use_protected_resource()
        pprint(f"Process {process_id} Leaving critical section", process_id)


if __name__ == "__main__":
    # Run
    with multiprocessing.Pool(4) as p:
        print("\n============================================================================")
        print("Lock demo: Only one process may be in the critical section at a time")
        print("============================================================================")
        p.map(access_exclusive_resource, range(4))

        print("\n============================================================================")
        print("Semaphore demo: Up to two processes may be in the critical section at a time")
        print("============================================================================")
        p.map(access_limited_resource, range(4))



from functools import lru_cache

from redis import Redis
from redis.asyncio import Redis as AsyncRedis


@lru_cache(maxsize=1)
def get_sync_redis(connection_url: str) -> Redis:
    """Get a synchronous Redis connection, which uses a connection pool"""
    return Redis.from_url(connection_url)


@lru_cache(maxsize=1)
def get_async_redis(connection_url: str) -> AsyncRedis:
    """Get an asynchronous Redis connection, which uses a connection pool"""
    return AsyncRedis.from_url(connection_url)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Hamilton Kibbe <ham@hamiltonkib.be>

"""Base classes for distributed synchronization primitives

"""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from types import TracebackType
from typing import Optional, Type


TUTTI_LOGGER_NAME = "tutti"


@dataclass(kw_only=True)
class LockConfig(ABC):
    """Lock congfiguration Base class.

    Parameters:
        name (str):
            The name of the Redis Lock.
        timeout (float):
            The timeout for acquiring the lock, in seconds
        blocking (bool):
            Whether to block while waiting for the lock to be released. Defaults to `True`.
    """
    name: str
    timeout: float
    blocking: bool = True


@dataclass
class SemaphoreConfig(ABC):
    """Semaphore congfiguration Base class.

    Parameters:
        name (str):
            The name of the Semaphore.
        max_concurrency (int):
            The maximum number of concurrent accesses to the resource.
    """
    name: str
    max_concurrency: int


class LockABC(ABC):
    """Abstract base class for Lock primitive

    Mostly compatible with :py:class:`threading.Lock`.
    """

    @abstractmethod
    def acquire(self, blocking: bool = True, timeout: Optional[float] = None) -> bool:
        """Acquire a lock.

        Parameters
        ----------
        blocking: bool
            Whether or not to block waiting for the lock. If blocking is set to `True`, block for up to `timeout`
            seconds (or forever if `timeout` is `None`).

        timeout: Optional[float]
            Number of seconds to wait for the lock if it is not immediately available. Defaults to `None`.

        Returns
        -------
        success : bool
            `True` if lock was acquired successfully, `False` otherwise.
        """
        pass

    @abstractmethod
    def release(self) -> None:
        """Release a lock.

        Returns
        -------
        None

        Raises
        ------
        `RuntimeError`
            if called on an unlocked Lock.
        """
        pass

    @abstractmethod
    def locked(self) -> bool:
        """Return `True` if the lock is locked

        Returns
        -------
        locked: bool
            `True` if the lock is held by anyone, not just this process.

        """
        pass

    @abstractmethod
    def __enter__(self) -> "LockABC":
        pass

    @abstractmethod
    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]) -> Optional[bool]:
        pass


class SemaphoreABC(ABC):
    """Abstract Base Class for Semaphore primitive

    Mostly compatible with :py:class:`threading.Semaphore`.
    """

    @abstractmethod
    def acquire(self, blocking: bool = True, timeout: Optional[float] = None) -> bool:
        pass

    @abstractmethod
    def release(self, n: int = 1) -> None:
        pass

    @abstractmethod
    def __enter__(self) -> "SemaphoreABC":
        pass

    @abstractmethod
    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]) -> Optional[bool]:
        pass


class AsyncLockABC(ABC):
    """Abstract base class for Lock primitive

    Mostly compatible with :py:class:`asyncio.Lock`.
    """

    @abstractmethod
    async def acquire(self, blocking: bool = True, timeout: Optional[float] = None) -> bool:
        """Acquire a lock.

        Parameters
        ----------
        blocking: bool
            Whether or not to block waiting for the lock. If blocking is set to `True`, block for up to `timeout`
            seconds (or forever if `timeout` is `None`).

        timeout: Optional[float]
            Number of seconds to wait for the lock if it is not immediately available. Defaults to `None`.

        Returns
        -------
        success : bool
            `True` if lock was acquired successfully, `False` otherwise.
        """
        pass

    @abstractmethod
    async def release(self) -> None:
        """Release a lock.

        Returns
        -------
        None

        Raises
        ------
        `RuntimeError`
            if called on an unlocked Lock.
        """
        pass

    @abstractmethod
    async def locked(self) -> bool:
        """Return `True` if the lock is locked

        Returns
        -------
        locked: bool
            `True` if the lock is held by anyone, not just this process.

        """
        pass

    @abstractmethod
    async def __aenter__(self) -> "AsyncLockABC":
        pass

    @abstractmethod
    async def __aexit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]) -> Optional[bool]:
        pass


class AsyncSemaphoreABC(ABC):
    """Abstract Base Class for Semaphore primitive

    Mostly compatible with :py:class:`asyncio.Semaphore`.
    """

    @abstractmethod
    async def acquire(self, blocking: bool = True, timeout: Optional[float] = None) -> bool:
        pass

    @abstractmethod
    async def release(self) -> None:
        pass

    @abstractmethod
    async def __aenter__(self) -> "AsyncSemaphoreABC":
        pass

    @abstractmethod
    async def __aexit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]) -> Optional[bool]:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Hamilton Kibbe <ham@hamiltonkib.be

from .sync import Lock, Semaphore, BoundedSemaphore


__version__ = '1.0.0'
__all__ = ["Lock", "Semaphore", "BoundedSemaphore"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Hamilton Kibbe <ham@hamiltonkib.be>

"""Redis distributed synchronization primitive backend"""

import logging

from redis import Redis
from redis.lock import Lock as RedisLock

from tutti.base import LockABC, SemaphoreABC, TUTTI_LOGGER_NAME

from .connection import get_sync_redis
from .utils import (
    acquire_lock,
    release_lock,
    acquire_semaphore,
    release_semaphore,
    locked,
)
from .types import RedisSemaphoreHandle


logger = logging.getLogger(TUTTI_LOGGER_NAME)


class RedisWrapper:
    @staticmethod
    def acquire_lock(
        conn: Redis,
        lock_name: str,
        blocking: bool = True,
        timeout: float | None = None
    ) -> tuple[RedisLock, bool]:
        return acquire_lock(conn, lock_name, blocking, timeout)

    @staticmethod
    def release_lock(conn: Redis, lock: RedisLock) -> bool:
        return release_lock(conn, lock)

    @staticmethod
    def locked(conn: Redis, lock_name: str) -> bool:
        return locked(conn, lock_name)

    @staticmethod
    def acquire_semaphore(
        conn: Redis,
        lock_name: str,
        value: int = 1,
        blocking: bool = True,
        timeout: float = -1
    ) -> RedisSemaphoreHandle | None:
        return acquire_semaphore(conn, lock_name, value, blocking, timeout)

    @staticmethod
    def release_semaphore(conn: Redis, lock: RedisSemaphoreHandle) -> bool:
        return release_semaphore(conn, lock)


class Lock(LockABC):

    def __init__(
        self,
        lock_name: str,
        connection_url: str,
        timeout: float,
        blocking: bool = True,
        conn: Redis | None = None,
        redis_wrapper: type[RedisWrapper] = RedisWrapper
    ) -> None:
        self._handle: RedisLock | None = None
        self._blocking = blocking
        self._timeout = timeout
        self._lock_name = lock_name
        self._connection_url = connection_url
        self._redis_wrapper = redis_wrapper

        if conn is not None and isinstance(conn, Redis):
            self._conn = conn
        else:
            self._conn = get_sync_redis(connection_url)

    def acquire(self, blocking: bool = True, timeout: float | None = None) -> bool:
        try:
            lock, result = self._redis_wrapper.acquire_lock(
                self._conn,
                lock_name=self._lock_name,
                blocking=blocking,
                timeout=self._get_timeout(timeout)
            )
        except Exception as e:
            logger.error(f"Error acquiring tutti lock: {e}", exc_info=True)
            self._handle = None
            return False
        else:
            if result:
                self._handle = lock
            return result

    def release(self) -> None:
        if self._handle is None:
            raise RuntimeError("Attempt to release unlocked lock.")
        self._redis_wrapper.release_lock(self._conn, self._handle)

    def locked(self) -> bool:
        return self._redis_wrapper.locked(self._conn, self._lock_name)

    def __enter__(self) -> "Lock":
        acquired = self.acquire(self._blocking, self._timeout)
        if not acquired:
            raise RuntimeError("Unable to acquire lock")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        return self.release()

    def _get_timeout(self, timeout: float | None = None) -> float | None:
        return self._timeout if self._timeout is not None else timeout


class Semaphore(SemaphoreABC):
    def __init__(
        self,
        lock_name: str,
        value: int,
        timeout: float,
        connection_url: str,
        conn: Redis | None = None,
        redis_wrapper: type[RedisWrapper] = RedisWrapper
    ) -> None:
        self._value = value
        self._handle: RedisSemaphoreHandle | None = None
        self._lock_name = lock_name
        self._timeout = timeout
        self._connection_url = connection_url
        self._redis_wrapper = redis_wrapper

        if conn is not None and isinstance(conn, Redis):
            self._conn = conn
        else:
            self._conn = get_sync_redis(connection_url)

    def acquire(self, blocking: bool = True, timeout: float | None = None) -> bool:
        timeout_float = self._get_timeout(timeout)
        lock_name = f"{self._lock_name}-lock"
        with Lock(
            lock_name=lock_name,
            blocking=blocking,
            connection_url=self._connection_url,
            timeout=timeout_float,
            conn=self._conn,
            redis_wrapper=self._redis_wrapper
        ):
            self._handle = self._redis_wrapper.acquire_semaphore(
                self._conn,
                value=self._value,
                lock_name=self._lock_name,
                blocking=blocking,
                timeout=self._timeout
            )
            return self._handle is not None

    def release(self, n: int = 1) -> None:
        if self._handle is not None:
            self._redis_wrapper.release_semaphore(self._conn, self._handle)

    def __enter__(self) -> "Semaphore":
        acquired = self.acquire()
        if not acquired:
            raise RuntimeError("Unable to acquire semaphore")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self._handle:
            return self.release()

    def _get_timeout(self, timeout: float | None = None) -> float:
        return self._timeout if self._timeout is not None else timeout if timeout is not None else -1


class BoundedSemaphore(Semaphore):
    def release(self, n: int = 1) -> None:
        if self._handle is None or not self._redis_wrapper.release_semaphore(self._conn, self._handle):
            raise ValueError("Semaphore released too many times")


__all__ = ["Lock", "Semaphore", "BoundedSemaphore"]



from .sync import Lock
from .sync import Semaphore
from .sync import BoundedSemaphore
from .asyncio import Lock as AsyncLock
from .asyncio import Semaphore as AsyncSemaphore
from .asyncio import BoundedSemaphore as AsyncBoundedSemaphore
from .types import RedisLockConfig
from .types import RedisSemaphoreConfig


__all__ = [
    "Lock",
    "Semaphore",
    "BoundedSemaphore",
    "AsyncLock",
    "AsyncSemaphore",
    "AsyncBoundedSemaphore",
    "RedisLockConfig",
    "RedisSemaphoreConfig",
]



from .base import LockABC, SemaphoreABC, LockConfig, SemaphoreConfig

from .backends.redis import (
    Lock as RedisLock,
    Semaphore as RedisSemaphore,
    BoundedSemaphore as RedisBoundedSemaphore,
    RedisLockConfig,
    RedisSemaphoreConfig,
)


class Lock:
    """Factory class for creating lock instances.
    This class is a factory for creating lock instances based on the provided
    configuration. It uses the `__new__` method to create an instance of the
    appropriate lock backend.

    Parameters
    ----------
    config: LockConfig
        The configuration object for the lock. This should be an instance of
        a subclass of LockConfig, such as RedisLockConfig.

    Raises
    ------
    NotImplementedError
        If the backend is not implemented.

    Returns
    -------
    LockABC
        An instance of the appropriate lock backend. The type of the lock
        backend is determined by the configuration object passed to the
        factory
    """
    def __new__(cls, config: LockConfig) -> LockABC:
        return _lock_factory(config)


class Semaphore:
    """Factory class for creating semaphore instances.
    This class is a factory for creating semaphore instances based on the provided
    configuration. It uses the `__new__` method to create an instance of the
    appropriate semaphore backend.

    Parameters
    ----------
    config: SemaphoreConfig
        The configuration object for the semaphore. This should be an instance of
        a subclass of SemaphoreConfig, such as RedisSemaphoreConfig.

    Raises
    ------
    NotImplementedError
        If the backend is not implemented.

    Returns
    -------
    SemaphoreABC
        An instance of the appropriate semaphore backend. The type of the semaphore
        backend is determined by the configuration object passed to the
        factory
    """
    def __new__(cls, config: SemaphoreConfig) -> SemaphoreABC:
        return _semaphore_factory(config)


class BoundedSemaphore:
    """Factory class for creating bounded semaphore instances.
    This class is a factory for creating semaphore instances based on the provided
    configuration. It uses the `__new__` method to create an instance of the
    appropriate semaphore backend.

    Parameters
    ----------
    config: SemaphoreConfig
        The configuration object for the semaphore. This should be an instance of
        a subclass of SemaphoreConfig, such as RedisSemaphoreConfig.

    Raises
    ------
    NotImplementedError
        If the backend is not implemented.

    Returns
    -------
    SemaphoreABC
        An instance of the appropriate semaphore backend. The type of the semaphore
        backend is determined by the configuration object passed to the
        factory
    """
    def __new__(cls, config: SemaphoreConfig) -> SemaphoreABC:
        return _bounded_semaphore_factory(config)


def _lock_factory(config: LockConfig) -> LockABC:
    """Factory function to create an Lock instance based on the configuration.
    Args:
        config (LockConfig): The tutti lock configuration object.
    Raises:
        NotImplementedError: If the backend is not implemented.
    Returns:
        LockABC: An instance of the appropriate Lock backend.
    """
    match config:
        case RedisLockConfig():
            return RedisLock(
                connection_url=config.connection_url,
                lock_name=config.name,
                timeout=config.timeout,
                blocking=config.blocking,
            )
        case _:
            raise NotImplementedError(f"The {config} backend is not implemented for sync locks.")


def _semaphore_factory(config: SemaphoreConfig) -> SemaphoreABC:
    """Factory function to create an Semaphore instance based on the configuration.
    Args:
        config (SemaphoreConfig): The tutti semaphore configuration object.
    Raises:
        NotImplementedError: If the backend is not implemented.
    Returns:
        SemaphoreABC: An instance of the appropriate Semaphore backend.
    """
    match config:
        case RedisSemaphoreConfig():
            return RedisSemaphore(
                connection_url=config.connection_url,
                value=config.max_concurrency,
                lock_name=config.name,
                timeout=config.lock_timeout,
            )
        case _:
            raise NotImplementedError(f"The {config} backend is not implemented for sync semaphores.")


def _bounded_semaphore_factory(config: SemaphoreConfig) -> SemaphoreABC:
    """Factory function to create an BoundedSemaphore instance based on the configuration.
    Args:
        config (SemaphoreConfig): The tutti semaphore configuration object.
    Raises:
        NotImplementedError: If the backend is not implemented.
    Returns:
        SemaphoreABC: An instance of the appropriate BoundedSemaphore backend.
    """
    match config:
        case RedisSemaphoreConfig():
            return RedisBoundedSemaphore(
                connection_url=config.connection_url,
                value=config.max_concurrency,
                lock_name=config.name,
                timeout=config.lock_timeout,
            )
        case _:
            raise NotImplementedError(f"The {config} backend is not implemented for sync bounded semaphores.")


__all__ = ["Lock", "Semaphore", "BoundedSemaphore"]
